from datetime import datetime


def datetime2rfc3339(dt, is_utc=False):
    if is_utc is False:
        # calculating timezone
        d1 = datetime.now()
        d2 = datetime.utcnow()
        diff_hr = (d1 - d2).seconds / 60 / 60
        tz = ""

        if diff_hr == 0:
            tz = "Z"
        else:
            if diff_hr > 0:
                tz = "+%s" % (tz)

            tz = "%s%.2d%.2d" % (tz, diff_hr, 0)

        return "%s%s" % (dt.strftime("%Y-%m-%dT%H:%M:%S.%f"), tz)

    else:
        return dt.isoformat() + 'Z'



# List of syslog facilities as per RFC5424
FAC_KERNEL = 0
FAC_USER = 1
FAC_MAIL = 2
FAC_SYSTEM = 3
FAC_SECURITY = 4
FAC_SYSLOG = 5
FAC_PRINTER = 6
FAC_NETWORK = 7
FAC_UUCP = 8
FAC_CLOCK = 9
FAC_AUTH = 10
FAC_FTP = 11
FAC_NTP = 12
FAC_LOG_AUDIT = 13
FAC_LOG_ALERT = 14
FAC_CLOCK2 = 15
FAC_LOCAL0 = 16
FAC_LOCAL1 = 17
FAC_LOCAL2 = 18
FAC_LOCAL3 = 19
FAC_LOCAL4 = 20
FAC_LOCAL5 = 21
FAC_LOCAL6 = 22
FAC_LOCAL7 = 23

# List of syslog severities as per RFC5424
SEV_EMERGENCY = 0
SEV_ALERT = 1
SEV_CRITICAL = 2
SEV_ERROR = 3
SEV_WARNING = 4
SEV_NOTICE = 5
SEV_INFO = 6
SEV_DEBUG = 7



import os
import uuid

class TempFile:
    def __init__(self, data=None, path="/tmp"):
        """
        Initialize a TempFile object.
        
        :param data: Content to be written to the file (optional)
        :param path: Directory where the temporary file will be created (default: /tmp)
        """
        self.path = None
        self._data = data
        self._base_path = path
        self._file_created = False

    def save(self):
        """
        Create the file with the specified data.
        Generates a unique filename using UUID.
        
        :return: Full path to the created file
        """
        # Generate a unique filename using UUID
        filename = str(uuid.uuid4())
        self.path = os.path.join(self._base_path, filename)
        
        # Ensure the base directory exists
        os.makedirs(self._base_path, exist_ok=True)
        
        # Write data to the file if provided
        if self._data is not None:
            with open(self.path, 'w') as f:
                f.write(str(self._data))
        
        self._file_created = True
        return self.path

    def delete(self):
        """
        Delete the temporary file if it exists.
        """
        if self._file_created and self.path and os.path.exists(self.path):
            os.unlink(self.path)
            self._file_created = False
            self.path = None

    def __enter__(self):
        """
        Context manager entry point.
        Creates the file when entering the context.
        
        :return: Self
        """
        self.save()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Context manager exit point.
        Deletes the file when exiting the context.
        """
        self.delete()

    @property
    def data(self):
        """
        Getter for the file data.
        
        :return: Content of the file
        """
        if self.path and os.path.exists(self.path):
            with open(self.path, 'r') as f:
                return f.read()
        return self._data


import asyncio
import logging
import socket
import ssl
from datetime import datetime
from .const import FAC_USER, SEV_INFO
from .helpers import datetime2rfc3339
from .tmpfile import TempFile
from . import exceptions as exc

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger()


class SyslogClient:
    def __init__(
        self,
        server: str,
        port: int,
        proto: str = 'UDP',
        forceipv4: bool = False,
        clientname: str = None,
        rfc: str = None,
        maxMessageLength: int = 2048,
        cert_data: dict = dict(),
        timeout: int = 30
    ) -> None:
        self.socket = None
        self.server = server
        self.port = port
        self.proto = proto
        self.rfc = rfc
        self.maxMessageLength = maxMessageLength
        self.forceipv4 = forceipv4
        self.clientname=clientname
        self.use_tls = True if proto.upper() == 'TLS' else False
        self.ssl_context = None
        self.timeout = timeout
        if self.use_tls:
            self.cafile = cert_data['cafile']
            self.certfile = cert_data.get('certfile')
            self.keyfile = cert_data.get('keyfile')

        if self.clientname is None:
            self.clientname = socket.getfqdn() or socket.gethostname() or "aiosyslog-client"

    async def send(self, message: bytes):
        message = message[: self.maxMessageLength]
        if self.proto.upper() == "UDP":
            res = await self._send_udp(message)
            return res
        elif self.proto.upper() in ["TCP", "TLS"]:
            res = await self._send_tcp(message)
            return res
        else:
            raise ValueError("Unsupported protocol. Use 'udp','tcp' or 'tls'")

    async def _send_udp(self, message):
        loop = asyncio.get_running_loop()
        transport, _ = await loop.create_datagram_endpoint(
            lambda: asyncio.DatagramProtocol(), remote_addr=(self.server, self.port)
        )
        transport.sendto(message)
        transport.close()

    def get_ssl_context(self):
        ssl_context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
        try:
            # can pass the content of the file directly. kinda hacky, but whatever.
            if self.cafile.startswith("-----BEGIN"):
                ssl_context.load_verify_locations(cadata=self.cafile)
            else:
                ssl_context.load_verify_locations(self.cafile)
        except Exception as ex:
            raise exc.ServerCertificateLoadError(f"Could not load server certificate: {ex}")
        # if using client certificate authentication
        if self.certfile and self.keyfile:
            try:
                if self.certfile.startswith("-----BEGIN") and self.keyfile.startswith("-----BEGIN"):
                        # this is kinda not effective at all, ngl.. but i dont wanna store keys on the drive.
                        with TempFile(self.certfile) as cert, TempFile(self.keyfile) as key:
                            ssl_context.load_cert_chain(certfile=cert.path, keyfile=key.path)
                else:
                    ssl_context.load_cert_chain(certfile=self.certfile, keyfile=self.keyfile)
            except Exception as ex:
                raise exc.ClientCertificateLoadError(f"Could not load client certificate: {ex}")
        return ssl_context

    # TODO: add a timeout decorator?
    async def _send_tcp(self, message):
        if self.use_tls and self.ssl_context is None:
            self.ssl_context=self.get_ssl_context()
        try:
            _, conn = await asyncio.wait_for(
                asyncio.open_connection(self.server, self.port, ssl=self.ssl_context),
                timeout=self.timeout
            )
            conn.write(message)
            await conn.drain()
            conn.close()
            await conn.wait_closed()
        except asyncio.TimeoutError:
            raise exc.SyslogConnectionTimeout(f"Timed out waiting, server={self.server}:{self.port}")
        except OSError as ex:
            e_str=str(ex)
            if "Errno 61" in e_str:
                raise exc.SyslogConnectionFailure(f"Connection refused, server={self.server}:{self.port}")
            elif "Errno 8" in e_str:
                raise exc.SyslogConnectionFailure(f"Invalid server name provided, server={self.server}:{self.port}")
            else:
                print(f"unknown OSError exception: {ex}")
        except Exception as ex:
            raise exc.UnknownSyslogResponseError(f"uknown error while sending message: {ex}")


class SyslogClientRFC5424(SyslogClient):
    def __init__(
        self,
        server: str,
        port: int,
        proto: str = 'udp',
        forceipv4: bool = False,
        clientname: str = None,
        cert_data: dict = dict(),
        timeout: int = 30
    ) -> None:
        super().__init__(
            server=server,
            port=port,
            proto=proto,
            forceipv4=forceipv4,
            clientname=clientname,
            rfc='5424',
            maxMessageLength=4096,
            cert_data=cert_data,
            timeout=timeout
        )

    async def log(
        self,
        message: str,
        facility: int = FAC_USER,
        severity: int = SEV_INFO,
        timestamp: datetime = None,
        hostname: str = None,
        version: int = 1,
        program: str = None,
        pid: int = None,
        msgid: int = None,
    ):
        pri = facility * 8 + severity
        timestamp_s = (
            datetime2rfc3339(datetime.utcnow(), is_utc=True)
            if timestamp is None
            else datetime2rfc3339(timestamp, is_utc=False)
        )
        # wtf is this for..
        hostname_s = self.clientname if hostname is None else hostname
        appname_s = "-" if program is None else program
        procid_s = "-" if pid is None else pid
        msgid_s = "-" if msgid is None else msgid

        formatted_payload = "<%i>%i %s %s %s %s %s %s\n" % (
            pri,
            version,
            timestamp_s,
            hostname_s,
            appname_s,
            procid_s,
            msgid_s,
            message,
        )
        response = await self.send(formatted_payload.encode('utf-8'))
        return response


class SyslogClientRFC3164(SyslogClient):
    def __init__(
        self,
        server: str,
        port: int,
        proto: str = 'udp',
        forceipv4: bool = False,
        clientname: str = None,
        cert_data: dict = dict(),
        timeout: int = 30
    ) -> None:
        super().__init__(
            server=server,
            port=port,
            proto=proto,
            forceipv4=forceipv4,
            clientname=clientname,
            rfc='3164',
            maxMessageLength=2048,
            cert_data=cert_data,
            timeout=timeout
        )

    async def log(
        self,
        message: str,
        facility: int = FAC_USER,
        severity: int = SEV_INFO,
        timestamp: datetime = datetime.now(),
        hostname: str = None,
        program: str = "SyslogClient",
        pid: int = None,
    ) -> None:
        pri = facility * 8 + severity
        timestamp_s = timestamp.strftime("%b %d %H:%M:%S")
        hostname_s = self.clientname if hostname is None else hostname

        if pid is not None:
            program += "[%i]" % (pid)

        d = "<%i>%s %s %s: %s\n" % (pri, timestamp_s, hostname_s, program, message)

        response = await self.send(d.encode('ASCII', 'ignore'))
        return response


if __name__ == '__main__':
    import doctest

    doctest.testmod()



class UnknownSyslogResponseError(Exception):
    pass

class SyslogConnectionFailure(Exception):
    pass

class SyslogConnectionTimeout(Exception):
    pass

class ClientCertificateLoadError(Exception):
    pass

class ServerCertificateLoadError(Exception):
    pass





from .syslog_client import SyslogClientRFC3164, SyslogClientRFC5424
from .tmpfile import TempFile
from . import exceptions

__all__ = ["SyslogClientRFC5424", "SyslogClientRFC3164", "TempFile", "exceptions"]



import argparse
import asyncio
import os

from syslog_client import (
    FAC_SYSLOG,
    SEV_DEBUG,
    SyslogClientRFC3164,
    SyslogClientRFC5424,
)


async def main():
    parser = argparse.ArgumentParser(description="AIO Syslog client")
    parser.add_argument("--server", "-s", type=str, help="Server name", required=True)
    parser.add_argument("--port", "-p", type=int, help="Port number", default=514)
    parser.add_argument(
        "--protocol", "-t", choices=["tcp", "udp", "tls"], help="Used transprot protocol", default="udp"
    )
    parser.add_argument("--rfc", "-r", choices=["5424", "3164"], help="RFC to use", default="5424")
    parser.add_argument("--program", "-o", type=str, help="Program name", default="SyslogClient")
    parser.add_argument("--pid", "-i", type=int, help="PID of program", default=os.getpid())
    parser.add_argument("--message", "-m", type=str, help="Message to send", required=True)
    parser.add_argument("--cafile", type=str, help="Server CA certificate", required=False, default="server.crt")
    parser.add_argument("--certfile", type=str, help="Client Certificate", required=False, default="client.crt")
    parser.add_argument("--keyfile", type=str, help="Client Private Key", required=False, default="client.key")

    args = parser.parse_args()
    cert_data = None
    if args.protocol.upper() == 'TLS':
        cert_data = dict(
            cafile=args.cafile,
            certfile=args.certfile,
            keyfile=args.keyfile,
        )

    if args.rfc == "5424":
        client = SyslogClientRFC5424(server=args.server, port=args.port, proto=args.protocol, cert_data=cert_data)
    else:
        client = SyslogClientRFC3164(server=args.server, port=args.port, proto=args.protocol, cert_data=cert_data)

    _ = await client.log(args.message, facility=FAC_SYSLOG, severity=SEV_DEBUG, program=args.program, pid=args.pid)


if __name__ == "__main__":
    asyncio.run(main())
